from cum import exceptions
import threading
import click
import json
import os
import re
import requests
import sys


class BaseConfig(object):
    def __init__(self):
        self.load()

    def __setattr__(self, name, value):
        """Ensures that changes made after loading with default values
        are written back to disk.
        """
        if hasattr(self, 'persistent_config'):
            self.persistent_config[name] = value
        object.__setattr__(self, name, value)

    @property
    def default_download_directory(self):
        """Returns a platform-specific download directory to use if no download
        directory is specified by the user.
        """
        if sys.platform in ['cygwin', 'win32']:
            return os.path.join(os.environ['USERPROFILE'], 'Downloads')
        else:
            return os.environ['HOME']

    def load(self):
        try:
            f = open(config_path)
        except FileNotFoundError:
            j = {}
        else:
            j = json.load(f)
            f.close()

        self.batoto = BatotoConfig(self, j.get('batoto', {}))
        self.cbz = j.get('cbz', False)
        self.compact_new = j.get('compact_new', False)
        self.download_directory = j.get('download_directory',
                                        self.default_download_directory)
        self.download_threads = j.get('download_threads', 4)
        self.html_parser = j.get('html_parser', 'html.parser')
        self.madokami = MadokamiConfig(self, j.get('madokami', {}))
        self.relative_latest = j.get('relative_latest', False)

        self.persistent_config = j

    def serialize(self):
        """Returns the current persistent configuration as a dictionary. All
        private configuration values starting with an underscore are removed
        from the configuration.
        """
        configuration = dict(self.persistent_config)
        configuration['batoto'] = dict(self.batoto.__dict__)
        configuration['madokami'] = dict(self.madokami.__dict__)
        configuration_keys = list(configuration.keys())
        while True:
            if not configuration_keys:
                break

            key = configuration_keys.pop(0)
            key_levels = key.split('.')
            dictionary = None
            value = configuration[key_levels[0]]
            for level in key_levels[1:]:
                dictionary = value
                value = value[level]

            if key_levels[-1].startswith('_'):
                del dictionary[key_levels[-1]]
                continue
            if isinstance(value, dict):
                dict_keys = ['.'.join([key, x]) for x in value]
                configuration_keys += dict_keys
        return configuration

    def write(self):
        if hasattr(self, 'persistent_config'):
            configuration = self.serialize()

            with open(config_path, 'w') as file:
                json.dump(configuration, file, sort_keys=True, indent=2)


class BatotoConfig(object):
    forum_url = 'https://bato.to/forums/index.php'
    login_query = {'app': 'core', 'module': 'global',
                   'section': 'login', 'do': 'process'}

    def __init__(self, config, dict):
        self._config = config
        self._lock = threading.Lock()
        self._login_attempts = 0
        self.cookie = dict.get('cookie', None)
        self.member_id = dict.get('member_id', None)
        self.pass_hash = dict.get('pass_hash', None)
        self.password = dict.get('password', None)
        self.username = dict.get('username', None)

    def login(self):
        self._login_attempts += 1
        if self._login_attempts > 1:
            raise exceptions.LoginError('Batoto login error')
        if not self.username:
            username = click.prompt('Batoto username')
        else:
            username = self.username
        if not self.password:
            password = click.prompt('Batoto password', hide_input=True)
        else:
            password = self.password
        r = requests.get(self.forum_url, params=self.login_query)
        auth_key = re.search(r"'auth_key' value='(.+)'", r.text).group(1)
        data = {'auth_key': auth_key,
                'referer': 'http://bato.to/',
                'ips_username': username,
                'ips_password': password,
                'rememberMe': 1}
        r = requests.post(self.forum_url, params=self.login_query, data=data)
        try:
            self.cookie = r.cookies['session_id']
            self.member_id = r.cookies['member_id']
            self.pass_hash = r.cookies['pass_hash']
        except KeyError:
            raise exceptions.LoginError('Batoto login error')
        self._config.write()

    @property
    def login_cookies(self):
        self._lock.acquire()
        if not (self.cookie and self.member_id and self.pass_hash):
            self.login()
        self._lock.release()
        return {'session_id': self.cookie,
                'member_id': self.member_id,
                'pass_hash': self.pass_hash}


class MadokamiConfig(object):
    def __init__(self, config, dict):
        self._config = config
        self.password = dict.get('password', None)
        self.username = dict.get('username', None)

    @property
    def login(self):
        """Returns a tuple containing the username and password. Missing values
        will be prompted from the user during runtime.
        """
        if not self.username:
            self.username = click.prompt('Madokami username')
        if not self.password:
            self.password = click.prompt('Madokami password', hide_input=True)
        return (self.username, self.password)


def get():
    """Returns the active config object."""
    return _config


def initialize(directory=None):
    """Initializes the cum directory and config file either with specified
    directory or ~/.cum.
    """
    global _config, config_path, cum_dir
    if directory:
        cum_dir = directory
    elif sys.platform in ['cygwin', 'win32']:
        cum_dir = os.path.join(os.environ['APPDATA'], 'cum')
    else:
        cum_dir = os.path.join(os.environ['HOME'], '.cum')
    if not os.path.exists(cum_dir):
        os.mkdir(cum_dir)
    config_path = os.path.join(cum_dir, 'config.json')
    _config = BaseConfig()



from bs4 import BeautifulSoup
from cum import config, exceptions, output
from cum.scrapers.base import BaseChapter, BaseSeries, download_pool
from functools import partial
from mimetypes import guess_type
import concurrent.futures
import re
import requests


class BatotoSeries(BaseSeries):
    url_re = re.compile(r'https?://bato\.to/comic/_/comics/')
    name_re = re.compile(r'Ch\.([A-Za-z0-9\.\-]*)(?:\: (.*))?')

    def __init__(self, url, **kwargs):
        super().__init__(url, **kwargs)
        self._get_page()
        self.chapters = self.get_chapters()

    def _get_page(self):
        r = requests.get(self.url, cookies=config.get().batoto.login_cookies)
        self.soup = BeautifulSoup(r.text, config.get().html_parser)

    def get_chapters(self):
        # Loops through as many times necessary until either registeration
        # notice is no longer found on the series page or login attempt limit
        # is reached, at which point LoginError is raised by config.
        while True:
            if self.soup.find('div', id='register_notice'):
                config.get().batoto.login()
                self._get_page()
            else:
                break

        rows = self.soup.find_all('tr', class_="row lang_English chapter_row")
        chapters = []
        for row in rows:
            columns = row.find_all('td')

            name = columns[0].img.next_sibling.strip()
            name_parts = re.search(self.name_re, name)
            chapter = name_parts.group(1)
            title = name_parts.group(2)

            url = columns[0].find('a').get('href')

            groups = [g.string for g in columns[2].find_all('a')]

            c = BatotoChapter(name=self.name, alias=self.alias,
                              chapter=chapter, url=url,
                              groups=groups, title=title)
            chapters.append(c)
        return chapters

    @property
    def name(self):
        return self.soup.find('h1').string.strip()


class BatotoChapter(BaseChapter):
    error_re = re.compile(r'ERROR \[10010\]')
    img_path_re = re.compile(
        r'(http://img.bato.to/comics/.*?/img)([0-9]*)(\.[A-Za-z]+)'
    )
    ch_img_path_re = re.compile(
        r'(http://img.bato.to/comics/.*?/ch[0-9]*p)([0-9]*)(\.[A-Za-z]+)'
    )
    next_img_path_re = re.compile(r'img.src = "(.+)";')
    url_re = re.compile(r'https?://bato.to/reader#(.*)')
    uses_pages = True

    @staticmethod
    def _reader_get(chapter_hash, page_index):
        return requests.get('http://bato.to/areader',
                            params={'id': chapter_hash, 'p': page_index},
                            headers={'Referer': 'http://bato.to/reader'},
                            cookies=config.get().batoto.login_cookies)

    def available(self):
        if not self.batoto_hash:
            return False
        self.r = self.reader_get(1)
        if not len(self.r.text):
            return False
        elif self.r.status_code == 404:
            return False
        elif "The thing you're looking for is unavailable." in self.r.text:
            if "Try logging in" not in self.r.text:
                return False
            else:
                return True
        else:
            return True

    @property
    def batoto_hash(self):
        hash_match = re.search(self.url_re, self.url)
        if hash_match:
            return hash_match.group(1)
        else:
            return None

    def download(self):
        if getattr(self, 'r', None):
            r = self.r
        else:
            r = self.reader_get(1)
        soup = BeautifulSoup(r.text, config.get().html_parser)
        if soup.find('a', href='#{}_1_t'.format(self.batoto_hash)):
            # The chapter uses webtoon layout, meaning all of the images are on
            # the same page.
            pages = [''.join(i) for i in re.findall(self.img_path_re, r.text)]
        else:
            pages = [x.get('value') for x in
                     soup.find('select', id='page_select').find_all('option')]
            if not pages:
                output.warning('Skipping {s.name} {s.chapter}: '
                               'chapter has no pages'
                               .format(s=self))
                return
            # Replace the first URL with the first image URL to avoid scraping
            # the first page twice.
            pages[0] = soup.find('img', id='comic_page').get('src')
            next_match = re.search(self.next_img_path_re, r.text)
            if next_match:
                pages[1] = next_match.group(1)
        files = [None] * len(pages)
        futures = []
        last_image = None
        with self.progress_bar(pages) as bar:
            for i, page in enumerate(pages):
                try:
                    if guess_type(page)[0]:
                        image = page
                    else:   # Predict the next URL based on the last URL
                        for reg in [self.img_path_re, self.ch_img_path_re]:
                            m = re.match(reg, last_image)
                            if m:
                                break
                        else:
                            raise ValueError
                        mg = list(m.groups())
                        num_digits = len(mg[1])
                        mg[1] = "{0:0>{digits}}".format(int(mg[1]) + 1,
                                                        digits=num_digits)
                        image = "".join(mg)
                    r = requests.get(image, stream=True)
                    if r.status_code == 404:
                        r.close()
                        raise ValueError
                except ValueError:  # If we fail to do prediction, scrape
                    r = self.reader_get(i + 1)
                    soup = BeautifulSoup(r.text, config.get().html_parser)
                    image = soup.find('img', id='comic_page').get('src')
                    image2_match = re.search(self.next_img_path_re, r.text)
                    if image2_match:
                        pages[i + 1] = image2_match.group(1)
                    r = requests.get(image, stream=True)
                fut = download_pool.submit(self.page_download_task, i, r)
                fut.add_done_callback(partial(self.page_download_finish,
                                              bar, files))
                futures.append(fut)
                last_image = image
            concurrent.futures.wait(futures)
            self.create_zip(files)

    def from_url(url):
        chapter_hash = re.search(BatotoChapter.url_re, url).group(1)
        r = BatotoChapter._reader_get(chapter_hash, 1)
        soup = BeautifulSoup(r.text, config.get().html_parser)
        try:
            series_url = soup.find('a', href=BatotoSeries.url_re)['href']
        except TypeError:
            raise exceptions.ScrapingError('Chapter has no parent series link')
        series = BatotoSeries(series_url)
        for chapter in series.chapters:
            if chapter.url == url:
                return chapter

    def reader_get(self, page_index):
        return self._reader_get(self.batoto_hash, page_index)



from abc import ABCMeta, abstractmethod
from alembic.migration import MigrationContext
from alembic.operations import Operations
from sqlalchemy import inspect
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.declarative.clsregistry import _ModuleMarker
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.sql import sqltypes


class DatabaseSanity(object):
    """Class designed for database validation and repair."""

    def __init__(self, base, engine):
        self.base = base
        self.engine = engine
        self.errors = []
        self.inspection_engine = inspect(self.engine)

    @property
    def database_tables(self):
        """Generator that returns table names from the database file."""
        for table_name in self.inspection_engine.get_table_names():
            yield table_name

    def find_database_column(self, table, column):
        """Looks for a column from the database file."""
        for engine_column in self.inspection_engine.get_columns(table):
            if engine_column['name'] == column.key:
                return engine_column

    @property
    def is_sane(self):
        """Returns a boolean value that indicates whether the database file
        passes the sanity checks or not based on the length of the error list.
        """
        return not len(self.errors)

    @property
    def model_tables(self):
        """Generator that returns tuples containing model name and model class.
        Models with classes inherited from _ModuleMarker class are skipped as
        they are not actual models.
        """
        for name, class_ in self.base._decl_class_registry.items():
            if isinstance(class_, _ModuleMarker):
                continue
            yield (name, class_)

    def test(self):
        """Runs through all the database sanity tests."""
        self.test_tables()
        for table in self.database_tables:
            self.test_columns(table)
        self.test_madokami_url()  # TODO: Remove in the future.

    def test_columns(self, table):
        """Tests the columns in database table. If a column exists in the
        table, check that the column has correct datatype. If the column
        doesn't exist, add a MissingColumn object to the error list.
        """
        columns = self.inspection_engine.get_columns(table)
        column_names = [c["name"] for c in columns]

        # Search for correct class from models for the mapper.
        mapper = None
        for name, class_ in self.model_tables:
            if class_.__tablename__ == table:
                mapper = inspect(class_)
                break
        if not mapper:
            return

        for column_property in mapper.attrs:
            if isinstance(column_property, RelationshipProperty):
                # TODO: Add RelationshipProperty sanity checking.
                pass
            else:
                for column in column_property.columns:
                    if column.key in column_names:
                        self.test_datatype(table, column)
                        self.test_nullable(table, column)
                    else:
                        self.errors.append(MissingColumn(table, column.key,
                                                         parent=self))

    def test_datatype(self, table, column):
        """Tests that database column datatype matches the one defined in the
        models.
        """
        database_column = self.find_database_column(table, column)

        if isinstance(column.type, sqltypes.String):
            expected_type = sqltypes.VARCHAR
        elif isinstance(column.type, sqltypes.Integer):
            expected_type = sqltypes.INTEGER
        elif isinstance(column.type, sqltypes.Boolean):
            expected_type = sqltypes.BOOLEAN
        elif isinstance(column.type, sqltypes.DateTime):
            expected_type = sqltypes.DATETIME

        if not isinstance(database_column['type'], expected_type):
            self.errors.append(
                DatatypeMismatch(table, database_column, expected_type,
                                 parent=self)
            )

    def test_madokami_url(self):
        """Temporary function to test if Madokami series and chapters have the
        correct URL after the top-level domain change. Will be removed after an
        appropriate time has passed for people have made the switch.
        """
        global db
        from cum import db
        old_domain = 'manga.madokami.com'
        new_domain = 'manga.madokami.al'
        models = [db.Series, db.Chapter]
        for model in models:
            condition = model.url.ilike('%{}%'.format(old_domain))
            try:
                results = db.session.query(model).filter(condition).all()
            except SQLAlchemyError:
                continue
            if results:
                error = IncorrectDomain(model, old_domain, new_domain)
                self.errors.append(error)

    def test_nullable(self, table, column):
        """Tests that database column nullable status matches the one defined
        in the models.
        """
        database_column = self.find_database_column(table, column)

        if column.nullable != database_column['nullable']:
            error = NullableMismatch(table, column.key, column.nullable,
                                     parent=self)
            self.errors.append(error)

    def test_tables(self):
        """Tests if all the defined models are found in the database. If a
        table is missing, a MissingTable object is added to the error list.
        """
        for name, class_ in self.model_tables:
            table = class_.__tablename__
            if table not in self.database_tables:
                self.errors.append(MissingTable(table, parent=self))


class SanityError(metaclass=ABCMeta):
    """Base class for all sanity errors."""

    @abstractmethod
    def __str__():
        """Returns an error message that can be displayed to the user."""
        raise NotImplementedError

    @abstractmethod
    def fix(self):
        """Does appropriate changes to the database to rectify the issue."""
        raise NotImplementedError


class DatatypeMismatch(SanityError):
    """Class used for columns where the database datatype is different from the
    model datatype.
    """

    def __init__(self, table, column, expected_type, parent=None):
        self.column = column
        self.expected_type = expected_type
        self.parent = parent
        self.table = table

    def __str__(self):
        return ("{}.{} column has inappropriate datatype {} (should be {})"
                .format(self.table, self.column['name'],
                        self.column['type'], self.expected_type.__name__))

    def fix(self):
        """Uses Alembic batch operations to alter the column datatype in the table.
        """
        context = MigrationContext.configure(self.parent.engine.connect())
        op = Operations(context)
        for table in self.parent.base.metadata.sorted_tables:
            if table.name == self.table:
                for column in table.columns:
                    if column.name == self.column['name']:
                        with op.batch_alter_table(table.name) as batch_op:
                            batch_op.alter_column(column.name,
                                                  type_=column.type)
                        return


class IncorrectDomain(SanityError):
    """Class used for database entries that have an outdated domain."""

    def __init__(self, model, old_domain, new_domain):
        self.model = model
        self.old_domain = old_domain
        self.new_domain = new_domain

    def __str__(self):
        return ('{s.model.__tablename__} has entries with incorrect domain '
                '({s.old_domain} -> {s.new_domain})'
                .format(s=self))

    def fix(self):
        condition = self.model.url.ilike('%{}%'.format(self.old_domain))
        results = db.session.query(self.model).filter(condition).all()
        for result in results:
            result.url = result.url.replace(self.old_domain, self.new_domain)
        db.session.commit()


class MissingColumn(SanityError):
    """Class used for columns that are missing from the database tables."""

    def __init__(self, table, name, parent=None):
        self.name = name
        self.parent = parent
        self.table = table

    def __str__(self):
        return ('{s.table}.{s.name} column is missing from database'
                .format(s=self))

    def fix(self):
        context = MigrationContext.configure(self.parent.engine.connect())
        op = Operations(context)
        for table in self.parent.base.metadata.sorted_tables:
            if table.name == self.table:
                for column in table.columns:
                    if column.name == self.name:
                        with op.batch_alter_table(table.name) as batch_op:
                            batch_op.add_column(column.copy())
                        return


class MissingTable(SanityError):
    """Class used for tables that are missing from the database file."""

    def __init__(self, name, parent=None):
        self.name = name
        self.parent = parent

    def __str__(self):
        return ('{} table is missing from database'
                .format(self.name))

    def fix(self):
        """Searches for the schema.Table object from SQLAlchemy metadata and
        creates it using the engine bind.
        """
        for table in self.parent.base.metadata.sorted_tables:
            if table.name == self.name:
                table.create(bind=self.parent.engine)
                break


class NullableMismatch(SanityError):
    """Class used for database columns where NULLABLE differs from the model.
    """

    def __init__(self, table, name, is_nullable, parent=None):
        self.table = table
        self.name = name
        self.is_nullable = is_nullable
        self.parent = parent

    def __str__(self):
        if self.is_nullable:
            msg = 'is not nullable'
        else:
            msg = 'is nullable'
        return '{s.table}.{s.name} {msg}'.format(s=self, msg=msg)

    def fix(self):
        context = MigrationContext.configure(self.parent.engine.connect())
        op = Operations(context)
        for table in self.parent.base.metadata.sorted_tables:
            if table.name == self.table:
                for column in table.columns:
                    if column.name == self.name:
                        with op.batch_alter_table(table.name) as batch_op:
                            batch_op.alter_column(column.name,
                                                  nullable=self.is_nullable)
                        return



from abc import ABCMeta, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from cum import config, db, output
from mimetypes import guess_extension
from re import match, sub
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound
from tempfile import NamedTemporaryFile
import click
import os
import requests
import sys
import zipfile


download_pool = ThreadPoolExecutor(config.get().download_threads)


class BaseSeries(metaclass=ABCMeta):
    """Class that is used to represent an individual series on a site."""

    def __init__(self, url, **kwargs):
        self.url = url
        self.directory = kwargs.get('directory', None)

    @property
    def alias(self):
        """Returns an alias version of the series name, which only allows a
        certain command-line friendly set of characters.
        """
        allowed_re = r'[A-Za-z0-9\-\s]'

        # Take the series name, lowercase it, replace all spaces with dashes
        # and then replaces all repeating dashes with a single dash.
        name = sub('-+', '-', self.name.lower().replace(' ', '-'))

        # Return the string where all characters are matched to allowed_re.
        return ''.join(c for c in name if match(allowed_re, c))

    def follow(self, ignore=False):
        """Adds the series details to database and all current chapters."""

        try:
            s = db.session.query(db.Series).filter_by(url=self.url).one()
        except NoResultFound:
            s = db.Series(self)
            s.check_alias_uniqueness()
            output.series('Adding follow for {s.name} ({s.alias})'.format(s=s))
            db.session.add(s)
            db.session.commit()
        else:
            if s.following:
                output.warning('You are already following {s.name} ({s.alias})'
                               .format(s=s))
            else:
                s.directory = self.directory
                s.following = True
                db.session.commit()

        for chapter in self.chapters:
            chapter.save(s, ignore=ignore)

    @abstractmethod
    def get_chapters(self):
        """Returns a list of objects that represent all of the series' chapters
        and are based on the Chapter class.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self):
        """Returns a string containing the title of the series."""
        raise NotImplementedError

    def update(self):
        """Iterates through the currently available chapters and saves them in
        the database.
        """
        s = db.session.query(db.Series).filter_by(url=self.url).one()

        for chapter in self.chapters:
            chapter.save(s)


class BaseChapter(metaclass=ABCMeta):
    """Class that is used to represent an individual download on a site."""

    def __init__(self, *args, **kwargs):
        self.name = kwargs.get('name')
        self.alias = kwargs.get('alias')
        self.chapter = kwargs.get('chapter')
        self.title = kwargs.get('title', None)
        self.url = kwargs.get('url')
        self.groups = kwargs.get('groups', None)
        self.directory = kwargs.get('directory', None)

    def _strip_unwanted_characters(self, path):
        """Strips unwanted characters from paths or filenames."""
        KEEP_CHARACTERS = [' ', '.', '-', '_', '[', ']', '/', "'"]
        if sys.platform in ['cygwin', 'win32']:
            KEEP_CHARACTERS += [':', '\\']
        path = ''.join([char for char in path if char.isalpha() or
                        char.isdigit() or char in KEEP_CHARACTERS]).rstrip()
        path = sub(' +', ' ', path)
        return path

    def _windows_name_directory(self, directory):
        """Perform additional sanitization to ensure that the directory name
        complies with Windows naming conventions.
        """
        directory = sub(r'\.*$', '', directory)
        return directory

    def available(self):
        """Checks if chapter URL returns HTTP 404 or not, and returns a boolean
        value based on it. Broken links are pruned from the database.

        Some sites might not return HTTP 404 on missing chapters and require
        custom version of this method to work.
        """
        r = requests.head(self.url)
        if r.status_code == 404:
            return False
        else:
            return True

    def create_directory(self, directory):
        """Attempts to create a directory based on the supplied argument. If
        the directory cannot be created, attempts to create an acceptable
        directory. Returns the path of the directory that has eventually been
        created.
        """
        if not os.path.exists(directory):
            try:
                os.makedirs(directory)
            except OSError as error:
                if error.errno == 22:
                    # Path is invalid, most likely due to Windows naming rules.
                    directory = self._windows_name_directory(directory)
                    if not os.path.exists(directory):
                        os.makedirs(directory)
                else:
                    raise error
        return directory

    def create_zip(self, files):
        """Takes a list of named temporary files, makes a ZIP out of them and
        closes the temporary files, deleting them. Files inside the .zip are
        organized based on the list order with rolling numbering padded to six
        digits and with the prefix 'image'.
        """
        with zipfile.ZipFile(self.filename, 'w') as z:
            for num, f in enumerate(files):
                root, ext = os.path.splitext(f.name)
                z.write(f.name, 'img{num:0>6}{ext}'.format(num=num, ext=ext))
                f.close()
                os.remove(f.name)

    def db_remove(self):
        """Removes the chapter from the database."""
        c = db.session.query(db.Chapter).filter_by(url=self.url).one()
        db.session.delete(c)
        try:
            db.session.commit()
        except:
            db.session.rollback()

    @abstractmethod
    def download(self):
        """Method that downloads the chapter and saves it in the appropriate
        directory as one archive file.
        """
        raise NotImplementedError

    @property
    def filename(self):
        name = self.name.replace('/', '')

        # Individually numbered chapter or a chapter range (e.g. '35',
        # '001-007').
        if match(r'[0-9\-]*$', self.chapter):
            chapter = ('c' +
                       '-'.join(x.zfill(3) for x in self.chapter.split('-')))
        # Individually numbered chapter with letter following the number
        # (e.g. '35v2', '9a').
        elif match(r'[0-9]*[A-Za-z][0-9]*?$', self.chapter):
            number = match(r'([0-9]*)[A-Za-z]', self.chapter).group(1)
            chapter = 'c{:0>3}'.format(number)
        # Individually numbered chapter with decimal (e.g. '1.5').
        elif match(r'[0-9]*\.[0-9]*$', self.chapter):
            number, decimal = self.chapter.split('.')
            chapter = 'c{:0>3} x{}'.format(number, decimal)
        # Failing all else, e.g. 'Special'. Becomes 'c000 [Special]'.
        else:
            chapter = 'c000 [{}]'.format(self.chapter)

        if self.groups:
            group = ''.join('[{}]'.format(g) for g in self.groups)
        else:
            group = '[Unknown]'

        if config.get().cbz:
            ext = 'cbz'
        else:
            ext = 'zip'

        if self.directory:
            directory = os.path.expanduser(self.directory)
        else:
            directory = name
        download_dir = os.path.expanduser(config.get().download_directory)
        download_dir = os.path.join(download_dir, directory)
        download_dir = self._strip_unwanted_characters(download_dir)
        download_dir = self.create_directory(download_dir)

        # Format the filename somewhat based on Daiz's manga naming scheme.
        # Remove any '/' characters to prevent the name of the manga splitting
        # the files into an unwanted sub-directory.
        filename = '{} - {} {}.{}'.format(name, chapter, group,
                                          ext).replace('/', '')
        filename = self._strip_unwanted_characters(filename)

        # Join the path parts and sanitize any unwanted characters that might
        # cause issues with filesystems. Remove repeating whitespaces.
        target = os.path.join(download_dir, filename)

        return target

    @staticmethod
    @abstractmethod
    def from_url(url):
        """Method to initialize a Chapter object from the chapter URL."""
        raise NotImplementedError

    def get(self, use_db=True):
        """Downloads the chapter if it is available.

        Optionally does not attempt to remove the chapter from the database or
        mark the chapter as downloaded if `db_remove` is set to False.
        """
        if self.available():
            self.download()
            if use_db:
                self.mark_downloaded()
        elif use_db:
            output.warning('Removing {} {}: missing from remote'
                           .format(self.name, self.chapter))
            self.db_remove()

    def ignore(self):
        """Fetches the chapter from the database and marks it ignored."""
        c = db.session.query(db.Chapter).filter_by(url=self.url).one()
        c.downloaded = -1
        db.session.commit()

    def mark_downloaded(self):
        """Fetches the chapter from the database and marks it downloaded."""
        c = db.session.query(db.Chapter).filter_by(url=self.url).one()
        c.downloaded = 1
        db.session.commit()

    def mark_new(self):
        """Fetches the chapter from the database and marks it new."""
        c = db.session.query(db.Chapter).filter_by(url=self.url).one()
        c.downloaded = 0
        db.session.commit()

    @staticmethod
    def page_download_finish(bar, files, fs):
        """Callback functions for page_download_task futures, assigning the
        resulting filehandles to the right index in the array and updating
        the progress bar.
        """
        index, f = fs.result()
        files[index] = f
        bar.update(1)

    @staticmethod
    def page_download_task(page_num, r):
        """Saves the response body of a single request, returning the file
        handle and the passed through number of the page to allow for non-
        sequential downloads in parallel.
        """
        ext = guess_extension(r.headers.get('content-type'))
        f = NamedTemporaryFile(suffix=ext, delete=False)
        for chunk in r.iter_content(chunk_size=4096):
            if chunk:
                f.write(chunk)
        f.flush()
        f.close()
        r.close()
        return((page_num, f))

    def progress_bar(self, arg):
        """Returns a pre-configured Click progress bar to use with downloads.
        If chapter uses separate page downloads, page download progress is
        shown (e.g. '7/20').
        """
        if self.uses_pages:
            iterable = arg
            length = None
        else:
            iterable = None
            length = arg

        click.echo('{c.alias} {c.chapter}'.format(c=self))
        return click.progressbar(iterable=iterable, length=length,
                                 fill_char='>', empty_char=' ',
                                 show_pos=self.uses_pages, show_percent=True)

    def save(self, series, ignore=False):
        """Save a chapter to database."""
        try:
            c = db.Chapter(self, series)
        except IntegrityError:
            db.session.rollback()
        else:
            if ignore:
                c.downloaded = -1
            db.session.add(c)
            try:
                db.session.commit()
            except IntegrityError:
                db.session.rollback()



from bs4 import BeautifulSoup
from cum import config
from cum.scrapers.base import BaseChapter, BaseSeries, download_pool
from functools import partial
from urllib.parse import urljoin
import concurrent.futures
import re
import requests

name_re = re.compile(r'Chapter ([0-9\.]+)(?:$|\: )(.*)')
fallback_re = re.compile(r'(.*?)(?:$|\: )(.*)')


class DynastyScansSeries(BaseSeries):
    url_re = re.compile(r'http://dynasty-scans\.com/series/')

    def __init__(self, url, **kwargs):
        super().__init__(url, **kwargs)
        r = requests.get(url)
        self.soup = BeautifulSoup(r.text, config.get().html_parser)
        self.chapters = self.get_chapters()

    def get_chapters(self):
        chapter_list = self.soup.find('dl', class_='chapter-list')
        links = chapter_list.find_all('a', class_='name')
        chapters = []
        for link in links:
            name_parts = re.search(name_re, link.string)
            if not name_parts:
                name_parts = re.search(fallback_re, link.string)
            chapter = name_parts.group(1)
            title = name_parts.group(2)
            url = urljoin(self.url, link.get('href'))
            c = DynastyScansChapter(name=self.name, alias=self.alias,
                                    chapter=chapter, url=url, title=title)
            chapters.append(c)
        return chapters

    @property
    def name(self):
        return self.soup.find('h2', class_='tag-title').contents[0].string


class DynastyScansChapter(BaseChapter):
    url_re = re.compile(r'http://dynasty-scans\.com/chapters/')
    uses_pages = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.groups:
            self.groups = self.get_groups()

    def download(self):
        r = requests.get(self.url)
        pages = re.findall(r'"image":"(.*?)"', r.text)
        files = [None] * len(pages)
        futures = []
        with self.progress_bar(pages) as bar:
            for i, page in enumerate(pages):
                r = requests.get(urljoin(self.url, page), stream=True)
                fut = download_pool.submit(self.page_download_task, i, r)
                fut.add_done_callback(partial(self.page_download_finish,
                                              bar, files))
                futures.append(fut)
            concurrent.futures.wait(futures)
            self.create_zip(files)

    def from_url(url):
        r = requests.get(url)
        soup = BeautifulSoup(r.text, config.get().html_parser)
        series_url = urljoin(url,
                             soup.find('h3', id='chapter-title').a['href'])
        try:
            series = DynastyScansSeries(series_url)
        except:
            name = soup.find('h3', id='chapter-title').b.text
            return DynastyScansChapter(name=name, chapter='0', url=url)
        else:
            for chapter in series.chapters:
                if chapter.url == url:
                    return chapter
        return None

    def get_groups(self):
        r = requests.get(self.url)
        soup = BeautifulSoup(r.text, config.get().html_parser)
        scanlators = soup.find('span', class_='scanlators')
        if scanlators:
            links = scanlators.find_all('a')
        else:
            links = []
        groups = []
        for link in links:
            r = requests.get(urljoin(self.url, link.get('href')))
            s = BeautifulSoup(r.text, config.get().html_parser)
            g = s.find('h2', class_='tag-title').b.string
            groups.append(g)
        return groups



from cum import config, output, sanity
from natsort import humansorted
from shutil import copyfile
from sqlalchemy import (
    Boolean,
    Column,
    create_engine,
    DateTime,
    ForeignKey,
    Integer,
    String,
    Table
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker
from sqlalchemy.orm.exc import NoResultFound
from urllib.parse import urlparse
import click
import datetime
import os
import sqlalchemy.engine.url

Base = declarative_base()

group_table = Table(
    'group_association', Base.metadata,
    Column('chapter_id', Integer, ForeignKey('chapters.id')),
    Column('group_id', Integer, ForeignKey('groups.id'))
)


class Series(Base):
    __tablename__ = 'series'

    id = Column(Integer, primary_key=True)

    name = Column(String)
    alias = Column(String, unique=True, nullable=False)
    url = Column(String, unique=True)
    following = Column(Boolean, default=True)
    directory = Column(String)

    chapters = relationship("Chapter", backref="series")

    def __init__(self, series):
        self.name = series.name
        self.alias = series.alias
        self.url = series.url
        self.directory = series.directory

    @staticmethod
    def alias_lookup(alias):
        """Returns a DB object for a series by alias name. Prints an error if
        an invalid alias is specified.
        """
        try:
            s = (session.query(Series)
                 .filter_by(alias=alias, following=True)
                 .one())
        except NoResultFound:
            output.error('Could not find alias "{}"'.format(alias))
            exit(1)
        else:
            return s

    def check_alias_uniqueness(self):
        """Check if the series alias is unique before initalizing the series object.
        """
        alias = self.alias
        count = 1
        while session.query(Series).filter_by(alias=alias).all():
            alias = '{}-{}'.format(self.alias, count)
            count += 1
        self.alias = alias

    @property
    def last_added(self):
        """Returns the last time a chapter has been added to the series."""
        updates = sorted([x for x in self.chapters if x.added_on is not None],
                         reverse=True, key=lambda x: x.added_on)
        if updates:
            return updates[0].added_on
        else:
            return None

    @property
    def ordered_chapters(self):
        return humansorted(self.chapters, key=lambda x: x.chapter)


class Chapter(Base):
    __tablename__ = 'chapters'

    id = Column(Integer, primary_key=True)
    series_id = Column(Integer, ForeignKey('series.id'))

    # Downloaded has the value -1 for ignored chapters, 0 for new chapters and
    # 1 for downloaded chapters.
    downloaded = Column(Integer, default=0)
    chapter = Column(String)
    url = Column(String, unique=True)
    title = Column(String)
    added_on = Column(DateTime)

    groups = relationship('Group', secondary=group_table, backref='chapters')

    def __init__(self, chapter, series):
        self.series = series
        self.chapter = chapter.chapter
        self.title = chapter.title
        self.url = chapter.url
        self.added_on = datetime.datetime.now()

        self.groups = []
        for group in chapter.groups:
            try:
                g = session.query(Group).filter(Group.name == group).one()
            except NoResultFound:
                g = Group(group)
                session.add(g)
                session.commit()
            self.groups.append(g)

    @staticmethod
    def find_new(alias=None):
        """Return a list of new chapters as Chapter objects and applies human
        sorting to it. Accepts an optional 'alias' argument, which will filter
        the query.
        """
        query = session.query(Chapter).join(Series).filter(Series.following)
        if alias:
            query = query.filter(Series.alias == alias)
        query = query.filter(Chapter.downloaded == 0).all()
        return humansorted([x.to_object() for x in query],
                           key=lambda x: x.chapter)

    @property
    def group_tag(self):
        """Return a joined string of chapter's groups enclosed in brackets."""
        return ''.join(['[{}]'.format(x.name) for x in self.groups])

    @property
    def status(self):
        """Return the chapter's downloaded status as a one character flag."""
        if self.downloaded == 0:
            return 'n'
        elif self.downloaded == -1:
            return 'i'
        else:
            return ' '

    def to_object(self):
        """Turns a database entry into a chapter object for the respective
        site by parsing the URL.
        """
        parse = urlparse(self.url)
        kwargs = {'name': self.series.name,
                  'alias': self.series.alias,
                  'chapter': self.chapter,
                  'url': self.url,
                  'groups': self.groups,
                  'directory': self.series.directory}
        if parse.netloc == 'bato.to':
            from cum.scrapers.batoto import BatotoChapter
            return BatotoChapter(**kwargs)
        elif parse.netloc == 'dynasty-scans.com':
            from cum.scrapers.dynastyscans import DynastyScansChapter
            return DynastyScansChapter(**kwargs)
        elif parse.netloc == 'manga.madokami.al':
            from cum.scrapers.madokami import MadokamiChapter
            return MadokamiChapter(**kwargs)
        else:
            return None


class Group(Base):
    __tablename__ = 'groups'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name


def backup_database():
    """Backs up the database file to a file called cum.db.bak."""
    db_path = os.path.join(config.cum_dir, 'cum.db')
    backup_path = os.path.join(config.cum_dir, 'cum.db.bak')
    copyfile(db_path, backup_path)


def initialize():
    global db_path, engine, session
    db_path = os.path.join(config.cum_dir, 'cum.db')
    db_url = sqlalchemy.engine.url.URL('sqlite', database=db_path)
    engine = create_engine(db_url)
    if not os.path.exists(db_path):
        Base.metadata.create_all(bind=engine)
    Session = sessionmaker(bind=engine)
    session = Session()


def test_database():
    """Runs a database sanity test."""
    sanity_tester = sanity.DatabaseSanity(Base, engine)
    sanity_tester.test()
    if sanity_tester.errors:
        for error in sanity_tester.errors:
            err_target, err_msg = str(error).split(' ', 1)
            message = ' '.join([click.style(err_target, bold=True), err_msg])
            output.warning(message)
        output.error('Database has failed sanity check; '
                     'run `cum repair-db` to repair database')
        exit(1)



import click
from math import ceil


def chapter(msg):
    click.echo(
        click.style('==> ', fg='green') + msg
    )


def configuration(dictionary):
    settings = configuration_flatten(dictionary)
    for setting, value in sorted(settings.items()):
        if value is not None:
            click.echo('{} = {}'.format(setting, value))


def configuration_flatten(dictionary, parent_key=None):
    items = []
    for key, value in dictionary.items():
        if parent_key:
            new_key = '{}.{}'.format(parent_key, key)
        else:
            new_key = key
        if isinstance(value, dict):
            items.extend(configuration_flatten(value, new_key).items())
        else:
            items.append((new_key, value))
    return dict(items)


def error(msg):
    click.echo(
        click.style('==> ', fg='red') + msg
    )


def even_columns(items, bold_first_column=False, separator_width=1):
    """Prints two columns with the second column padded so that it always
    begins on the same line, regardless of how long the first column is on the
    same line. Takes input as a list of two item tuples.
    """
    first_column_length = len(max([x[0] for x in items], key=len))
    for item in items:
        padding = first_column_length - len(item[0])
        separator = ' ' * separator_width
        first_column = click.style(item[0] + ' ' * padding,
                                   bold=bold_first_column)
        indent = len(item[0]) + padding + separator_width
        line = click.wrap_text(separator.join([first_column, item[1]]),
                               subsequent_indent=' ' * indent,
                               width=click.get_terminal_size()[0])
        click.echo(line)


def list(items):
    """Print out a list of items in columns much like `ls` in bash."""
    if items:
        width = click.get_terminal_size()[0]
        longest = len(max(items, key=len))
        columns = int(width // (longest + 0.5))
        rows = ceil(len(items) / columns)

        for row in range(rows):
            line = []
            for column in range(columns):
                i = rows * column + row
                if i >= len(items):
                    continue
                line.append(items[i].ljust(longest))
            click.echo(' '.join(line))


def series(msg):
    click.echo(
        click.style('==> ', fg='blue', bold=True) + msg
    )


def warning(msg):
    click.echo(
        click.style('==> ', fg='yellow', bold=True) + msg
    )



from cum.scrapers.batoto import BatotoChapter, BatotoSeries
from cum.scrapers.dynastyscans import DynastyScansChapter, DynastyScansSeries
from cum.scrapers.madokami import MadokamiChapter, MadokamiSeries

series_scrapers = [BatotoSeries, DynastyScansSeries, MadokamiSeries]
chapter_scrapers = [BatotoChapter, DynastyScansChapter, MadokamiChapter]



#!/usr/bin/env python3
from cum import config, exceptions, output, version
from functools import wraps
import click
import concurrent.futures


class CumGroup(click.Group):
    def command(self, check_db=True, *args, **kwargs):
        def decorator(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                if check_db:
                    db.test_database()
                return f(*args, **kwargs)
            return super(CumGroup, self).command(*args, **kwargs)(wrapper)
        return decorator


def edit_defaults():
    """Edits the Click command default values after initializing the config."""
    latest_command = cli.get_command(cli, 'latest')
    for param in latest_command.params:
        if param.human_readable_name == 'relative':
            param.default = config.get().relative_latest


@click.command(cls=CumGroup)
@click.option('--cum-directory',
              help='Directory used by cum to store application files.')
@click.version_option(version=version.__version__,
                      message=version.version_string())
def cli(cum_directory=None):
    global db, output, sanity, utility
    config.initialize(directory=cum_directory)
    from cum import db, output, sanity, utility
    db.initialize()
    edit_defaults()


@cli.command()
@click.argument('alias')
def chapters(alias):
    """List all chapters for a manga series.

    Chapter listing will contain the flag value for the chapter ('n' for new,
    'i' for ignored and blank for downloaded), the chapter identifier ("chapter
    number") and the possible chapter title and group.
    """
    s = db.Series.alias_lookup(alias)
    if s.chapters:
        click.secho('f  chapter  title [group]', bold=True)
        for chapter in s.ordered_chapters:
            name_len = click.get_terminal_size()[0] - 11
            name = '{} {}'.format(chapter.title, chapter.group_tag)[:name_len]
            row = '{}  {:>7}  {}'.format(chapter.status, chapter.chapter, name)
            if row[0] == 'n':
                style = {'fg': 'white', 'bold': True}
            elif row[0] == ' ':
                style = {'bold': True}
            else:
                style = {}
            click.secho(row, **style)


@cli.command(name='config')
@click.argument('mode')
@click.argument('setting', required=False)
@click.argument('value', required=False)
def config_command(mode, setting, value):
    """Get or set configuration options.

    Mode can be either "get" or "set", depending on whether you want to read or
    write configuration values. If mode is "get", you can specify a setting to
    read that particular setting or omit it to list out all the settings. If
    mode is "set", you must specify the setting to change and assign it a new
    value.
    """
    if mode == 'get':
        if setting:
            parameters = setting.split('.')
            value = config.get()
            for parameter in parameters:
                try:
                    value = getattr(value, parameter)
                except AttributeError:
                    output.error('Setting not found')
                    exit(1)
            output.configuration({setting: value})
        else:
            configuration = config.get().serialize()
            output.configuration(configuration)
    elif mode == 'set':
        if setting is None:
            output.error('You must specify a setting')
            exit(1)
        if value is None:
            output.error('You must specify a value')
            exit(1)
        parameters = setting.split('.')
        preference = config.get()
        for parameter in parameters[0:-1]:
            try:
                preference = getattr(preference, parameter)
            except AttributeError:
                output.error('Setting not found')
                exit(1)
        try:
            current_value = getattr(preference, parameters[-1])
        except AttributeError:
            output.error('Setting not found')
            exit(1)
        if current_value is not None:
            if isinstance(current_value, bool):
                if value.lower() == 'false' or value == 0:
                    value = False
                else:
                    value = True
            else:
                try:
                    value = type(current_value)(value)
                except ValueError:
                    output.error('Type mismatch: value should be {}'
                                 .format(type(current_value).__name__))
                    exit(1)
        setattr(preference, parameters[-1], value)
        config.get().write()
    else:
        output.error('Mode must be either get or set')
        exit(1)


@cli.command()
@click.argument('aliases', required=False, nargs=-1)
def download(aliases):
    """Download all available chapters.

    If an optional alias is specified, the command will only download new
    chapters for that alias.
    """
    chapters = []
    if not aliases:
        chapters = db.Chapter.find_new()
    for alias in aliases:
        chapters += db.Chapter.find_new(alias=alias)
    output.chapter('Downloading {} chapters'.format(len(chapters)))
    for chapter in chapters:
        try:
            chapter.get()
        except exceptions.LoginError as e:
            output.warning('Could not download {c.alias} {c.chapter}: {e}'
                           .format(c=chapter, e=e.message))
            continue


@cli.command()
@click.argument('alias')
@click.argument('setting')
@click.argument('value')
def edit(alias, setting, value):
    """Modify settings for a follow.

    The following settings can be edited: alias, directory.
    """
    series = db.Series.alias_lookup(alias)
    alias = series.alias
    if value.lower() == 'none' or value.lower() == '-':
        value = None
    if setting == 'alias':
        series.alias = value
    elif setting == 'directory':
        series.directory = value
    else:
        setting = click.style(setting, bold=True)
        output.error('Invalid setting {}'.format(setting))
        exit(1)

    if not value:
        value = click.style('none', dim=True)
    else:
        value = click.style(value, bold=True)
    try:
        db.session.commit()
    except exceptions.DatabaseIntegrityError:
        db.session.rollback()
        output.error('Illegal value {}'.format(value))
        exit(1)
    else:
        output.chapter('Changed {} for {} to {}'.format(setting, alias, value))


@cli.command()
@click.argument('urls', required=True, nargs=-1)
@click.option('--directory',
              help='Directory which download the series chapters into.')
@click.option('--download', is_flag=True,
              help='Downloads the chapters for the added follows.')
@click.option('--ignore', is_flag=True,
              help='Ignores the chapters for the added follows.')
def follow(urls, directory, download, ignore):
    """Follow a series."""
    chapters = []
    for url in urls:
        try:
            series = utility.series_by_url(url)
        except exceptions.ScrapingError:
            output.warning('Scraping error ({})'.format(url))
            continue
        except exceptions.LoginError as e:
            output.warning('{} ({})'.format(e.message, url))
            continue
        if not series:
            output.warning('Invalid URL "{}"'.format(url))
            continue
        series.directory = directory
        if ignore:
            series.follow(ignore=True)
            output.chapter('Ignoring {} chapters'.format(len(series.chapters)))
        else:
            series.follow()
            chapters += db.Chapter.find_new(alias=series.alias)

    if download:
        output.chapter('Downloading {} chapters'.format(len(chapters)))
        for chapter in chapters:
            try:
                chapter.get()
            except exceptions.LoginError as e:
                output.warning('Could not download {c.alias} {c.chapter}: {e}'
                               .format(c=chapter, e=e.message))
                continue


@cli.command()
def follows():
    """List all follows.

    Will list all of the active follows in the database as a list of aliases.
    To find out more information on an alias, use the info command.
    """
    query = (db.session.query(db.Series)
             .filter_by(following=True)
             .order_by(db.Series.alias)
             .all())
    output.list([x.alias for x in query])


@cli.command()
@click.argument('input', required=True, nargs=-1)
@click.option('--directory',
              help='Directory which download chapters into.')
def get(input, directory):
    """Download chapters by URL or by alias:chapter.

    The command accepts input as either the chapter of the URL or the
    alias:chapter combination (e.g. 'bakuon:11'), if the chapter is already
    found in the database through a follow. The command will not enter the
    downloads in the database in case of URLs and ignores downloaded status
    in case of alias:chapter, so it can be used to download one-shots that
    don't require follows or for redownloading already downloaded chapters.
    """
    chapter_list = []
    for i in input:
        try:
            series = utility.series_by_url(i)
        except exceptions.ScrapingError:
            output.warning('Scraping error ({})'.format(i))
            continue
        except exceptions.LoginError as e:
            output.warning('{} ({})'.format(e.message, i))
            continue
        if series:
            chapter_list += series.chapters
        try:
            chapter = utility.chapter_by_url(i)
        except exceptions.ScrapingError:
            output.warning('Scraping error ({})'.format(i))
            continue
        except exceptions.LoginError as e:
            output.warning('{} ({})'.format(e.message, i))
            continue
        if chapter:
            chapter_list.append(chapter)
        if not series or chapter:
            try:
                a, c = i.split(':')
            except ValueError:
                output.warning('Invalid selection "{}"'.format(i))
            else:
                chapters = (db.session.query(db.Chapter)
                            .join(db.Series)
                            .filter(db.Series.alias == a,
                                    db.Chapter.chapter == c)
                            .all())
                for chapter in chapters:
                    chapter_list.append(chapter.to_object())
    for chapter in chapter_list:
        chapter.directory = directory
        try:
            chapter.get(use_db=False)
        except exceptions.LoginError as e:
            output.warning('Could not download {c.alias} {c.chapter}: {e}'
                           .format(c=chapter, e=e.message))
            continue


@cli.command()
@click.argument('alias')
@click.argument('chapters', required=True, nargs=-1)
def ignore(alias, chapters):
    """Ignore chapters for a series.

    Enter one or more chapters after the alias to ignore them. Enter the
    chapter identifiers as they are listed when using the chapters command. To
    ignore all of the chapters for a particular series, use the word "all" in
    place of the chapters.
    """
    utility.set_ignored(True, alias, chapters)


@cli.command()
@click.argument('alias', required=False)
@click.option('--relative/--no-relative', default=False,
              help='Uses relative times instead of absolute times.')
def latest(alias, relative):
    """List most recent chapter addition for series."""
    query = db.session.query(db.Series)
    if alias:
        query = query.filter_by(following=True, alias=alias)
    else:
        query = query.filter_by(following=True)
    query = query.order_by(db.Series.alias).all()
    updates = []
    for series in query:
        if series.last_added is None:
            time = 'never'
        elif relative:
            time = utility.time_to_relative(series.last_added)
        else:
            time = series.last_added.strftime('%Y-%m-%d %H:%M')
        updates.append((series.alias, time))
    output.even_columns(updates, separator_width=3)


@cli.command()
def new():
    """List all new chapters."""
    utility.list_new()


@cli.command()
@click.argument('alias')
def open(alias):
    """Open the series URL in a browser."""
    s = db.Series.alias_lookup(alias)
    click.launch(s.url)


@cli.command(check_db=False, name='repair-db')
def repair_db():
    """Runs an automated database repair."""
    sanity_tester = sanity.DatabaseSanity(db.Base, db.engine)
    sanity_tester.test()
    if sanity_tester.errors:
        output.series('Backing up database to cum.db.bak')
        db.backup_database()
        output.series('Running database repair')
        for error in sanity_tester.errors:
            error.fix()


@cli.command()
@click.argument('alias')
def unfollow(alias):
    """Unfollow manga.

    Will mark a series as unfollowed. In order not to lose history of
    downloaded chapters, the series is merely marked as unfollowed in the
    database rather than removed.
    """
    s = db.Series.alias_lookup(alias)
    s.following = False
    db.session.commit()
    output.series('Removing follow for {}'.format(s.name))


@cli.command()
@click.argument('alias')
@click.argument('chapters', required=True, nargs=-1)
def unignore(alias, chapters):
    """Unignore chapters for a series.

    Enter one or more chapters after the alias to mark them as new. Enter the
    chapter identifiers as they are listed when using the chapters command. To
    unignore all of the chapters for a particular series, use the word "all" in
    place of the chapters.
    """
    utility.set_ignored(False, alias, chapters)


@cli.command()
def update():
    """Gather new chapters from followed series."""
    pool = concurrent.futures.ThreadPoolExecutor(config.get().download_threads)
    futures = []
    warnings = []
    aliases = {}
    query = db.session.query(db.Series).filter_by(following=True).all()
    output.series('Updating {} series'.format(len(query)))
    for follow in query:
        fut = pool.submit(utility.series_by_url, follow.url)
        futures.append(fut)
        aliases[fut] = follow.alias
    with click.progressbar(length=len(futures), show_pos=True,
                           fill_char='>', empty_char=' ') as bar:
        for future in concurrent.futures.as_completed(futures):
            try:
                series = future.result()
            except exceptions.ConnectionError:
                warnings.append('Unable to update {} (connection error)'
                                .format(aliases[future]))
            except exceptions.ScrapingError:
                warnings.append('Unable to update {} (scraping error)'
                                .format(aliases[future]))
            except exceptions.LoginError as e:
                warnings.append('Unable to update {} ({})'
                                .format(aliases[future], e.message))
            else:
                series.update()
            bar.update(1)
    for w in warnings:
        output.warning(w)
    utility.list_new()


if __name__ == '__main__':
    cli()



from bs4 import BeautifulSoup
from contextlib import closing
from cum import config, exceptions
from cum.scrapers.base import BaseChapter, BaseSeries
from urllib.parse import urljoin
import re
import requests

fallback_re = re.compile(r'\- (.*) (?:\[(.*)\])?')
name_re = re.compile(r'-? c([0-9-]+).*?(?: \[(.*)\])?\.')


class MadokamiSeries(BaseSeries):
    url_re = re.compile(r'https://manga\.madokami\.al/Manga/[^.]+$')

    def __init__(self, url, **kwargs):
        super().__init__(url, **kwargs)
        self.session = requests.Session()
        self.session.auth = requests.auth.HTTPBasicAuth(*config
                                                        .get().madokami.login)
        r = self.session.get(url)
        if r.status_code == 401:
            raise exceptions.LoginError('Madokami login error')
        self.soup = BeautifulSoup(r.text, config.get().html_parser)
        self.chapters = self.get_chapters()

    def get_chapters(self):
        try:
            rows = (self.soup
                    .find('table', class_='mobile-files-table')
                    .find_all('tr'))
        except AttributeError:
            raise exceptions.ScrapingError()
        chapters = []
        for row in rows[1:]:
            # If the Read link cannot be found in the current row, the row is
            # assumed to be a non-manga file uploaded to the directory and will
            # thus be skipped.
            if not row.find('a', text='Read'):
                continue

            link = row.find('a')

            url = urljoin(self.url, link.get('href'))

            name = link.string
            name_parts = re.search(name_re, name)
            if not name_parts:
                name_parts = re.search(fallback_re, name)
            try:
                chapter = name_parts.group(1)
            except:
                continue
            if name_parts.group(2):
                groups = name_parts.group(2).split('][')
            else:
                groups = []

            c = MadokamiChapter(name=self.name, alias=self.alias,
                                chapter=chapter, url=url, groups=groups,
                                session=self.session)
            chapters.append(c)
        return chapters

    @property
    def name(self):
        return self.soup.find('span', class_='title').string


class MadokamiChapter(BaseChapter):
    url_re = re.compile(r'https://manga\.madokami\.al/Manga/.*/.*/.*\..*')
    uses_pages = False

    def __init__(self, *args, **kwargs):
        self.session = kwargs.get("session", requests.Session())
        super().__init__(*args, **kwargs)

    def download(self):
        if not self.session.auth:
            self.session.auth = requests.auth.HTTPBasicAuth(*config
                                                            .get()
                                                            .madokami.login)
        with closing(self.session.get(self.url, stream=True)) as r:
            if r.status_code == 401:
                raise exceptions.LoginError('Madokami login error')
            total_length = r.headers.get('content-length')
            with open(self.filename, 'wb') as f:
                if total_length is None:
                    f.write(r.content)
                else:
                    total_length = int(total_length)
                    with self.progress_bar(total_length) as bar:
                        for chunk in r.iter_content(chunk_size=4096):
                            if chunk:
                                bar.update(len(chunk))
                                f.write(chunk)
                    f.flush()

    def from_url(url):
        series_url = re.search(r'(.*manga\.madokami\.al/.*/)', url).group(1)
        series = MadokamiSeries(series_url)
        for chapter in series.chapters:
            if chapter.url == url:
                return chapter
        return None



from setuptools import find_packages, setup
import os
import re
import subprocess
import sys


script_path = os.path.dirname(os.path.realpath(__file__))
git_path = os.path.join(script_path, '.git')
re_version = r'v([0-9\.]*-?[0-9]*)?'
COMMAND_DESCRIBE = ['git', 'describe', '--tags', '--exact-match']
COMMAND_DESCRIBE_VERSION = ['git', 'describe', '--tags']
COMMAND_REV_PARSE = ['git', 'rev-parse', '--short', 'HEAD']
SUBPROCESS_KWARGS = {
    'cwd': script_path,
    'stdout': subprocess.PIPE,
    'stderr': subprocess.PIPE
}


class PythonVersionException(Exception):
    __module__ = Exception.__module__

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return repr(self.message)


def get_setup_version():
    """Return a version string that can be used with the setup method. Includes
    additional commits since the last tagged commit.
    """
    if os.path.isdir('.git'):
        process = subprocess.Popen(COMMAND_DESCRIBE_VERSION,
                                   **SUBPROCESS_KWARGS)
        process.wait()
        version = process.communicate()[0].decode("utf-8").strip()
        return re.match(re_version, version).group(1)
    else:
        from cum.version import __version__
        return __version__


def get_version():
    """Returns a version that will either result as "0.1" style string if the
    git commit is tagged or as "git-ebe0338" otherwise. If there is no
    directory named ".git'", the function will instead return False.
    """
    version_is_commit = False
    if os.path.isdir('.git'):
        process = subprocess.Popen(COMMAND_DESCRIBE, **SUBPROCESS_KWARGS)
        process.wait()
        if process.returncode != 0:
            version_is_commit = True
            process = subprocess.Popen(COMMAND_REV_PARSE, **SUBPROCESS_KWARGS)
            process.wait()
        version = process.communicate()[0].decode("utf-8").strip()
        if version_is_commit:
            version = '-'.join(['git', version])
        else:
            version = version[1:]
        return version
    else:
        return False


def write_version_file():
    """Writes a version.py file containing the current version. Only replaces
    the first line of the file so that the rest of the variables and functions
    can still be used normally.
    """
    version_file_path = os.path.join(script_path, 'cum/version.py')
    with open(version_file_path, 'r') as file:
        lines = file.readlines()
        lines[0] = "__version__ = '{}'\n".format(version)
    with open(version_file_path, 'w') as file:
        for line in lines:
            file.write(line)


if sys.version_info[0] < 3:
    raise PythonVersionException('Python 3.x required.')


version = get_version()
if version:
    write_version_file()

setup(
    name='cum',
    version=get_setup_version(),
    description='comic updater, mangafied',
    url='https://github.com/Hamuko/cum',
    author='Hamuko',
    author_email='hamuko@burakku.com',
    license='Apache2',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP'
    ],
    packages=find_packages(),
    install_requires=[
        'alembic',
        'beautifulsoup4',
        'Click',
        'natsort',
        'requests',
        'SQLAlchemy'
    ],
    extras_require={
        'testing': ['codecov', 'cov-core', 'nose2', 'pycodestyle']
    },
    entry_points={
        'console_scripts': ['cum=cum.cum:cli'],
    }
)



from requests.exceptions import ConnectionError
from sqlalchemy.exc import IntegrityError as DatabaseIntegrityError


class CumException(Exception):
    """Base class for all cum exceptions."""

    def __init__(self, message=''):
        self.message = message

    def __str__(self):
        return repr(self.message)


class ScrapingError(CumException):
    pass


class LoginError(CumException):
    pass



from cum import db, config, output
from cum.scrapers import chapter_scrapers, series_scrapers
import click
import datetime
import re


def chapter_by_url(url):
    """Helper function that iterates through the chapter scrapers defined in
    cum.scrapers.__init__ and returns an initialized chapter object when it
    matches the URL regex.
    """
    for Chapter in chapter_scrapers:
        if re.match(Chapter.url_re, url):
            return Chapter.from_url(url)


def list_new():
    """Helper method used in multiple cum commands to print out the new chapter
    details for each series. Has two possible styles for displaying the
    information: compact and normal.
    """
    items = {}
    for chapter in db.Chapter.find_new():
        try:
            items[chapter.alias].append(chapter.chapter)
        except KeyError:
            items[chapter.alias] = [chapter.chapter]
    if not items:
        return

    # Create a sorted list of tuples where the first value of the tuple is the
    # series alias and second is a string containing the new chapters for the
    # series separated by two spaces.
    new = [(x, '  '.join([y for y in items[x]])) for x in sorted(items)]

    if config.get().compact_new:
        print_new_compact(new)
    else:
        print_new_normal(new)


def print_new_compact(items):
    """Prints the new chapter information. E.g.
        joukamachi-no-dandelion 30  31  32  33  34
        minami-ke               153  154  155  156  157
    """
    output.even_columns(items, bold_first_column=True)


def print_new_normal(items):
    """Prints the new chapter information. E.g.
        joukamachi-no-dandelion
        30  31  32  33  34
        minami-ke
        153  154  155  156  157
    """
    width = click.get_terminal_size()[0]
    for series in items:
        click.secho(series[0], bold=True)
        click.echo(click.wrap_text(series[1], width=width))


def series_by_url(url):
    """Helper function that iterates through the series scrapers defined in
    cum.scrapers.__init__ and returns an initialized series object when it
    matches the URL regex.
    """
    for Series in series_scrapers:
        if re.match(Series.url_re, url):
            return Series(url)


def set_ignored(mark_ignored, alias, chapters):
    """Helper function for `cum ignore` and `cum unignore` commands, which will
    either ignore chapters if mark_ignored is True or unignore chapters if
    mark_ignored is False.
    """
    if mark_ignored:
        downloaded = 0
        message_start = 'I'
        method = 'ignore'
    else:
        downloaded = -1
        message_start = 'Uni'
        method = 'mark_new'

    s = db.Series.alias_lookup(alias)
    query = (db.session.query(db.Chapter)
                       .filter(db.Chapter.series == s,
                               db.Chapter.downloaded == downloaded))
    if len(chapters) == 1 and chapters[0].lower() == 'all':
        click.echo('{}gnoring {} chapters for {}'
                   .format(message_start, len(s.chapters), s.name))
        click.confirm('Do you want to continue',
                      prompt_suffix='? ', abort=True)
    else:
        query = query.filter(db.Chapter.chapter.in_(chapters))

    chapters = [x.to_object() for x in query.all()]
    for chapter in chapters:
        function = getattr(chapter, method)
        function()
    if len(chapters) == 1:
        output.chapter('{}gnored chapter {} for {}'
                       .format(message_start, chapters[0].chapter, s.name))
    else:
        output.series('{}gnored {} chapters for {}'
                      .format(message_start, len(chapters), s.name))


def time_to_relative(time):
    """Converts a DateTime object into a string containing the relative time of
    the DateTime, e.g. "3 minutes ago".
    """
    delta = datetime.datetime.now() - time
    if delta.days >= 30:
        value = delta.days // 30
        unit = 'months'
    elif delta.days > 0:
        value = delta.days
        unit = 'days'
    elif delta.seconds >= 60 * 60:
        value = delta.seconds // (60 * 60)
        unit = 'hours'
    elif delta.seconds >= 60:
        value = delta.seconds // 60
        unit = 'minutes'
    else:
        value = delta.seconds
        unit = 'seconds'
    if value == 1:
        unit = unit.rstrip('s')
    return '{} {} ago'.format(value, unit)
